/**
 *  MPQSettings.java: Preferences handling for the MPQ app.
 *            Copyright (C) 2012,2013 donfede@casagrau.org
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * created 2012-10-18 donfede@casagrau.org
 * credit to: Android SDK 10 / SipDemo app
 *
 * $Id: MPQSettings.java,v 1.10 2013-04-21 18:01:03 donfede Exp $
 */

package org.casagrau.mpq.mpq_android;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.nio.channels.FileChannel;
import java.text.SimpleDateFormat;
import java.util.Date;

import org.casagrau.mpq.mpq_android.SessionsDBProvider.SessionsDB;

import android.content.SharedPreferences;
import android.database.Cursor;
import android.net.Uri;
import android.os.Bundle;
import android.os.Environment;
import android.preference.EditTextPreference;
import android.preference.Preference;
import android.preference.PreferenceActivity;
import android.preference.PreferenceManager;
import android.util.Log;
import android.widget.Toast;

/**
 * Handles MPQ general settings for the app. Pretty much only reads and process
 * the XML description of preferences.
 */
public class MPQSettings extends PreferenceActivity {
	private static final String LOGTAG = "MPQSettings";

	@SuppressWarnings("deprecation")
	@Override
	public void onCreate(Bundle savedInstanceState) {
		// Note that none of the preferences are actually defined here.
		// They're all in the XML file res/xml/preferences.xml.
		super.onCreate(savedInstanceState);
		addPreferencesFromResource(R.xml.preferences);
	}
	
	/**
	 * Customize preference display of SD card space.
	 */
	@SuppressWarnings("deprecation")
	@Override
	public void onResume() {
		super.onResume();
		
		// Associate listeners for sync/backup/restore.
		Preference lp;
		lp = (Preference) findPreference(getString(R.string.pref_backup_key)) ;
		lp.setOnPreferenceClickListener(clickerPrefBackup);
		lp = (Preference) findPreference(getString(R.string.pref_sync_key)) ;
		lp.setOnPreferenceClickListener(clickerPrefSync);
		
		// Update total audio file usage
		EditTextPreference letp = (EditTextPreference) findPreference(getString(R.string.pref_audiomax_key)) ;
		long lspace_total_mb = 0;
		long lspace_free_mb = 0;
		SharedPreferences lsharedPref = PreferenceManager.getDefaultSharedPreferences(this);
		String laudioDirPref = lsharedPref.getString(getString(R.string.pref_audiodirectory_key), getString(R.string.pref_audiodirectory_default));
		File lFile = new File(Environment.getExternalStorageDirectory().getAbsolutePath() + laudioDirPref);
		lspace_total_mb = lFile.getTotalSpace() / 1024 / 1024;
		lspace_free_mb = lFile.getUsableSpace() / 1024 / 1024;
		
		// Query total audio file usage.
        Uri luri = SessionsDB.CONTENT_PRACTICES_FILE_SUM_KB;
        Log.d(LOGTAG, "onResume() 07   uri: " + luri);
        long lfile_sum_kb = 0;
        
        Cursor lcursor = getContentResolver().query(luri, null, null, null, null);
		if ( (null != lcursor) && (1 == lcursor.getCount()) ) {
			Log.d(LOGTAG, "onResume() 12 cursor: " + lcursor.toString() );
			lfile_sum_kb = lcursor.getLong(0);
			Log.d(LOGTAG, "onResume() 13  file sum kb: " + lfile_sum_kb);
		}		
		long lfile_sum_mb = lfile_sum_kb / 1024;
		Log.d(LOGTAG, "onResume() 06  tot:" + lspace_total_mb + " usable: " + lspace_free_mb);

		letp.setSummary(getString(R.string.pref_audiomax_summary) + " " + lspace_total_mb + "/" + lspace_free_mb + "\nUsed by MPQ: " + lfile_sum_mb);
	}
	
	
	
	/**************************************************
	 * 
	 * Helper methods, that respond to user clicks.
	 */
	
	/**
	 * Sync MPQ database with Internet server.
	 */
	private Preference.OnPreferenceClickListener clickerPrefSync = new Preference.OnPreferenceClickListener() {
		public boolean onPreferenceClick(Preference preference) {
			Log.d(LOGTAG, "clickerPrefSync onClick() 00  ");
			
			// Sync database.
			try {
				
// FIXME: need to place a check, so only 1x sync can run at a time.
//				Executor le = new Executor();
//				new MPQSyncTask().execute(le.SERIAL_EXCUTOR);
				new MPQSyncTask(preference.getContext()).execute();
		    } catch (IllegalStateException e) {
		        Toast.makeText(preference.getContext(), e.toString(), Toast.LENGTH_LONG).show();
		    }
			Log.d(LOGTAG, "clickerPrefSync onClick() 99");
			return true;		
		}
	}; // END Preference.OnPreferenceClickListener clickerPrefSync



	/**
	 * Backup MPQ database to SD card.
	 */
	private Preference.OnPreferenceClickListener clickerPrefBackup = new Preference.OnPreferenceClickListener() {
		public boolean onPreferenceClick(Preference preference) {
			Log.d(LOGTAG, "clickerPrefBackup onClick() 00  ");
			
			// Copy/backup database.
			try {
				SharedPreferences lsharedPref = PreferenceManager.getDefaultSharedPreferences(preference.getContext());
				String laudioDirPref = lsharedPref.getString(getString(R.string.pref_audiodirectory_key), getString(R.string.pref_audiodirectory_default));
				File lFile = new File(Environment.getExternalStorageDirectory().getAbsolutePath() + laudioDirPref);

		    	Log.d(LOGTAG, "clickerPrefBackup onClick() 20 mAudioServiceFileName f: " + lFile);
		    	
				// Calculate db backup filename date-suffix, based on current date stamp.
		    	String ldbFileName = null;
	        	ldbFileName = lFile + File.separator + "meta_dir" + File.separator + "mpq_practices_";

	        	SimpleDateFormat lfileDateFormat = new SimpleDateFormat("yyyyMMdd_HHmm"); // FIXME: this should likely be moved to strings.xml; or, even better, a centralized MPQ formatting module. 
				long lNowSec = java.lang.System.currentTimeMillis();
	        	ldbFileName = ldbFileName + lfileDateFormat.format(new Date(lNowSec)) + ".db";
		    	Log.d(LOGTAG, "clickerPrefBackup onClick() 39 ldbFileName: " + ldbFileName);

		    	// Source database file.
		        File data = Environment.getDataDirectory();
		        // /data/data/org.casagrau.mpq.mpq_android/databases/mpq_practices.db
	            String currentDBPath = "//data//"+ "org.casagrau.mpq.mpq_android" +"//databases//"+"mpq_practices.db";	            
	            File currentDB = new File(data, currentDBPath);
	            
	            // Destination backup file.
	            File backupDB = new File(ldbFileName);

	            // Connect streams, and copy data to complete backup.
	            FileChannel src = new FileInputStream(currentDB).getChannel();
                FileChannel dst = new FileOutputStream(backupDB).getChannel();
                dst.transferFrom(src, 0, src.size());
                src.close();
                dst.close();
                Toast.makeText(preference.getContext(), "Backup OK. File:" + backupDB.toString(), Toast.LENGTH_LONG).show();
		    } catch (Exception e) {
		        Toast.makeText(preference.getContext(), e.toString(), Toast.LENGTH_LONG).show();
		    }
			
			Log.d(LOGTAG, "clickerPrefBackup onClick() 99");
			return true;		
		}
	}; // END Preference.OnPreferenceClickListener clickerPrefBackup

	
	
} // END class MPQSettings
