/**
 * SessionsDBProvider.java: Content Provider to access practice sessions data.
 *            Copyright (C) 2012,2013 donfede@casagrau.org
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * created 2012-10-13 donfede@casagrau.org
 * credit: to the Android SDK sample Content Provider apps, NotesList and SearchableDictionary 
 *
 * $Id: SessionsDBProvider.java,v 1.53 2013-04-25 03:53:41 donfede Exp $
 */
package org.casagrau.mpq.mpq_android;

import java.io.File;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.TimeZone;

import org.casagrau.mpq.audio_service.MPQAudioService;

import android.content.ContentProvider;
import android.content.ContentUris;
import android.content.ContentValues;
import android.content.Context;
import android.content.Intent;
import android.content.SharedPreferences;
import android.content.UriMatcher;
import android.database.Cursor;
import android.database.SQLException;
import android.database.sqlite.SQLiteDatabase;
import android.database.sqlite.SQLiteOpenHelper;
import android.database.sqlite.SQLiteQueryBuilder;
import android.net.ParseException;
import android.net.Uri;
import android.os.Bundle;
import android.os.Environment;
import android.preference.PreferenceManager;
import android.provider.BaseColumns;
import android.util.Log;

/**
 * Provides access to a database of practice sessions and related data. 
 * Data "tables" include: Practices, Goals, Instruments
 * each with their own attributes.
 */
public class SessionsDBProvider extends ContentProvider {
    private static final String LOGTAG = "SessionsDBProvider";

    // DB constants
    private static final String DB_NAME = "mpq_practices.db";
    private static final int DB_VERSION = 19; // Update onUpgrade() along with DB version and schema changes.
    
    // Uri and UriMatcher constants.
    public static final String AUTHORITY = "org.casagrau.mpq.mpq_android.SessionsDBProvider";

    private static final int URIM_PRACTICES = 0;
    private static final int URIM_PRACTICES_ID = 10;   
    private static final int URIM_PRACTICES_COUNT = 20;
    private static final int URIM_PRACTICES_PREV_ID = 30;
    private static final int URIM_PRACTICES_NEXT_ID = 40;
    private static final int URIM_PRACTICES_FILE = 50;
    private static final int URIM_PRACTICES_TOTALS = 60;
    private static final int URIM_PRACTICES_FILE_SUM_KB = 65;
    private static final int URIM_GOALS = 70;
    private static final int URIM_GOALS_ID = 80;
    private static final int URIM_INSTRUMENTS = 90; 
    private static final int URIM_INSTRUMENTS_ID = 100;
    private static final int URIM_SYNC = 110; 
    private static final UriMatcher sUriMatcher = buildUriMatcher();      // Creates a UriMatcher object.
    
    /**
     * Builds up a UriMatcher for various table queries.
     */
    private static UriMatcher buildUriMatcher() {
        UriMatcher matcher =  new UriMatcher(UriMatcher.NO_MATCH);
        // to get definitions...
        matcher.addURI(AUTHORITY, SessionsDB.PRACTICES_TABLE, URIM_PRACTICES);
        matcher.addURI(AUTHORITY, SessionsDB.PRACTICES_TABLE+"/#", URIM_PRACTICES_ID);
        matcher.addURI(AUTHORITY, SessionsDB.PRACTICES_META_COUNT, URIM_PRACTICES_COUNT);
        matcher.addURI(AUTHORITY, SessionsDB.PRACTICES_META_PREV+"/#", URIM_PRACTICES_PREV_ID);
        matcher.addURI(AUTHORITY, SessionsDB.PRACTICES_META_NEXT+"/#", URIM_PRACTICES_NEXT_ID);
        matcher.addURI(AUTHORITY, SessionsDB.PRACTICES_META_FILE, URIM_PRACTICES_FILE);
        matcher.addURI(AUTHORITY, SessionsDB.PRACTICES_META_TOTALS, URIM_PRACTICES_TOTALS);
        matcher.addURI(AUTHORITY, SessionsDB.PRACTICES_META_FILE_SUM_KB, URIM_PRACTICES_FILE_SUM_KB);
        matcher.addURI(AUTHORITY, SessionsDB.GOALS_TABLE, URIM_GOALS);
        matcher.addURI(AUTHORITY, SessionsDB.GOALS_TABLE+"/#", URIM_GOALS_ID);
        matcher.addURI(AUTHORITY, SessionsDB.INSTRUMENTS_TABLE, URIM_INSTRUMENTS);
        matcher.addURI(AUTHORITY, SessionsDB.INSTRUMENTS_TABLE+"/#", URIM_INSTRUMENTS_ID);
        matcher.addURI(AUTHORITY, SessionsDB.SYNC_TABLE, URIM_SYNC);
        return matcher;
    }

    
    //////////////////////////////////////////////
    // public (inner)CONTRACT CLASS DEFINING TABLE NAMES AND COLUMNS 
    /**
     * Contract Class defining elements to access the Sessions database, including URI, table name, and table column constants.
     * This class is a Contract Class only - it cannot be instantiated.
     */
    public static final class SessionsDB implements BaseColumns{
    	//This class cannot be instantiated (it's meant to be a "Contract Class" for the content provider.
    	private SessionsDB() {}

    	/**
    	 * Practice table name
    	 */
        public static final String PRACTICES_TABLE = "practices";
    	
        
        /**
         * Practices table count - a meta query, returning -- select count(*).
         */
        public static final String PRACTICES_META_COUNT = "practices_count";
        
        /**
         * Practices table previous row _ID - a meta query, returning the previous available row ID.
         */
        public static final String PRACTICES_META_PREV = "practices_prev";
        
        /**
         * Practices table next row _ID - a meta query, returning the next available row ID.
         */
        public static final String PRACTICES_META_NEXT = "practices_next";
        
        /**
         * Practices table file to id - a meta query, returning the row _ID that corresponds with a filename.
         */
        public static final String PRACTICES_META_FILE = "practices_file";

        /**
         * Practices table totals for length and noise - a meta query, totals for columns.
         */
        public static final String PRACTICES_META_TOTALS = "practices_totals";

        /**
         * Practices table, sum of audio file SD space in KB - a meta query, sum of column.
         */
        public static final String PRACTICES_META_FILE_SUM_KB = "practices_file_sum_kb";
             
        /**
         * Practice date time stamp
         * <P>Type: INTEGER (Time in milliseconds)</P>
         */
        public static final String PRACTICE_DATE_STAMP = "practice_date_stamp";
        
        /**
         * Practice length total
         * <P>Type: INTEGER (seconds)
         */
        public static final String PRACTICE_LENGTH_TOTAL = "practice_length_total";
        
        /**
         * Practice length noise
         * <P>Type: INTEGER (seconds)
         */
         public static final String PRACTICE_LENGTH_NOISE = "practice_length_noise";
         
        /**
         * Practice audio file name
         * <P>Type: TEXT</P>
         */
         public static final String PRACTICE_AUDIO_FILE = "practice_audio_file";
         
        /**
         * Practice audio file size in KB (total file size)
         * <P>Type: INTEGER (KB)</P>
         */
         public static final String PRACTICE_AUDIO_FILE_KB = "practice_audio_file_KB";        

        /**
         * Practice goal id, FK.
         * <P>Type: INTEGER</P>
         */
        public static final String PRACTICE_GOAL_ID = "practice_goal_id";

        /**
         * Practice notes
         * <P>Type: TEXT</P>
         */
        public static final String PRACTICE_NOTES = "practice_notes";

        /**
         * Practice instrument id, FK.
         * <P>Type: INTEGER</P>
         */
        public static final String PRACTICE_INSTRUMENT_ID = "practice_instrument_id";
    	
        /**
         * Practice data version, used for sync.
         * <P>Type: INTEGER  (last edit time in milliseconds)</P>
         */
        public static final String PRACTICE_DATA_VER = "practice_data_ver";
        
    	/**
    	 * Goal table name
    	 */
        public static final String GOALS_TABLE = "goals";
    	
        /**
         * Goal name
         * <P>Type: TEXT</P>
         */
        public static final String GOAL_NAME = "goal_name";

        /**
         * Goal description
         * <P>Type: TEXT</P>
         */
        public static final String GOAL_DESC = "goal_desc";
        
        /**
         * Goal data version, used for sync.
         * <P>Type: INTEGER  (last edit time in milliseconds)</P>
         */
        public static final String GOAL_DATA_VER = "goal_data_ver";
        
        /**
         * Instrument table name
         */
        public static final String INSTRUMENTS_TABLE = "instruments";
        
        /**
         * Instrument name
         */
        public static final String INSTRUMENT_NAME = "inst_name";
        
        /**
         * Instrument description
         */
        public static final String INSTRUMENT_DESC = "inst_desc";

        /**
         * Instrument data version, used for sync.
         * <P>Type: INTEGER  (last edit time in milliseconds)</P>
         */
        public static final String INSTRUMENT_DATA_VER = "inst_data_ver";
        
        /**
         * Sync table name
         */
        public static final String SYNC_TABLE = "sync";
        
        /**
         * Userid name
         * <P>Type: TEXT</P>
         */
        public static final String SYNC_USERID = "sync_userid";
        
        /**
         * Userid pw
         * <P>Type: TEXT</P>
         */
        public static final String SYNC_USERPW = "sync_userpw";
        
        /**
         * Device id (ESN/MEID)
         * <P>Type: TEXT</P>
         */
        public static final String SYNC_DEVICE = "sync_device";
        
        /**
         * Database version.
         * <P>Type: INTEGER (matches SessionsDBProvider DB_VERSION)</P>
         */
        public static final String SYNC_DBVER = "sync_dbver";

        /**
         * MPQ version - string
         * <P>Type: TEXT (matches AndroidManifest android:versionName)</P>
         */
        public static final String SYNC_MPQVER = "sync_mpqver";

        /**
         * MPQ version - integer
         * <P>Type: INTEGER (matches AndroidManifest android:versionCode)</P>
         */
        public static final String SYNC_MPQVER_INT = "sync_mpqver_int";
        
        /**
         * Date stamp of last successful sync, in milliseconds.
         * <P>Type: INTEGER (Time in milliseconds)</P>
         */
        public static final String SYNC_DATE_STAMP = "sync_date_stamp";

        
        /**
         * Sync_crud table name
         */
        public static final String SYNC_CRUD_TABLE = "sync_crud";
        
        /**
         * CRUD operation, code letter.
         * <P>Type: TEXT (C create, U update, D delete)</P>
         */
        public static final String SYNC_CRUD_OP = "sync_crud_op";
        
        /**
         * CRUD object; a string with the table name (or URI? FIXME: verify, and update comment) receiving the operation.
         * <P>Type: TEXT (table name)</P>
         */
        public static final String SYNC_CRUD_OBJECT = "sync_crud_object";
        
        /**
         * CRUD id for the operation.
         * <P>Type: INTEGER</P>
         */
        public static final String SYNC_CRUD_ID = "sync_crud_id";
                
        
        /**
         * URI Constants
         */
    	public static final Uri CONTENT_PRACTICES_URI = Uri.parse("content://" + AUTHORITY + "/" + PRACTICES_TABLE);
    	public static final Uri CONTENT_PRACTICES_COUNT_URI = Uri.parse("content://" + AUTHORITY + "/" + PRACTICES_META_COUNT);
    	public static final Uri CONTENT_PRACTICES_PREV_URI = Uri.parse("content://" + AUTHORITY + "/" + PRACTICES_META_PREV);
    	public static final Uri CONTENT_PRACTICES_NEXT_URI = Uri.parse("content://" + AUTHORITY + "/" + PRACTICES_META_NEXT);    	
    	public static final Uri CONTENT_PRACTICES_FILE_URI = Uri.parse("content://" + AUTHORITY + "/" + PRACTICES_META_FILE);
    	public static final Uri CONTENT_PRACTICES_TOTALS_URI = Uri.parse("content://" + AUTHORITY + "/" + PRACTICES_META_TOTALS);
    	public static final Uri CONTENT_PRACTICES_FILE_SUM_KB = Uri.parse("content://" + AUTHORITY + "/" + PRACTICES_META_FILE_SUM_KB);
    	public static final Uri CONTENT_GOALS_URI = Uri.parse("content://" + AUTHORITY + "/" + GOALS_TABLE);  
        public static final Uri CONTENT_INSTRUMENTS_URI = Uri.parse("content://" + AUTHORITY + "/" + INSTRUMENTS_TABLE);
    	public static final Uri CONTENT_SYNC_URI = Uri.parse("content://" + AUTHORITY + "/" + SYNC_TABLE);  
        
        
        /**
         * Static helper methods, to provide consistency (ex: with data formatting)
         */

    	// ///////////////////////////////////////////////////
    	// Helper methods.
    	/**
    	 * Convert date in long int milliseconds format into ISO 8601 format.
    	 */
    	public static String longDateString(long dateSec) {
    		//Log.d(LOGTAG, "longDateStringLoader() 00 dateSet: " + dateSec);
    		String lStr = "";
    		SimpleDateFormat iso8601Format = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss"); 		// FIXME: this method will get called frequently -- Formater maybe should be static, or at the object level
    		try {
    			lStr = lStr + iso8601Format.format(new Date(dateSec));						
    		} catch (ParseException e) {
    			Log.e(LOGTAG, "longDateString - Parsing ISO8601 datetime failed", e);
    		}
    		return lStr;
    	} // END longDateString()
    	
    	/**
    	 * Convert session in milliseconds format into HH:mm:ss format. 
    	 */
    	public static String longSessionMSecString(long sessionMSec) {
    		//Log.d(LOGTAG, "longSessionMSecString() 00 sessionSec: " + sessionMSec);
    		String lStr = "";
    		
    		/** // requires API level >= 9
//    		lStr = String.format("%d min, %d sec", 
//    			    TimeUnit.MILLISECONDS.toMinutes(sessionMSec),
//    			    TimeUnit.MILLISECONDS.toSeconds(sessionMSec) - 
//    			    TimeUnit.MINUTES.toSeconds(TimeUnit.MILLISECONDS.toMinutes(sessionMSec))
//    			); 
    		 */	
    		int lseconds = (int) (sessionMSec / 1000) % 60 ;
    		int lminutes = (int) ((sessionMSec / (1000*60)) % 60);
    		int lhours   = (int) ((sessionMSec / (1000*60*60)) % 24);
    		lStr = String.format("%02d:%02d:%02d",
    				lhours, lminutes, lseconds);				
    		//Log.d(LOGTAG, "longSessionMSecString() 99       lStr: " + lStr);
    		return lStr;
    	} // END intSessionSecString()       
    	
       	/**
    	 * Convert session in milliseconds format into int minutes. 
    	 */
    	public static int longSessionMinInt(long sessionMSec) {
    		//Log.d(LOGTAG, "longSessionMinInt() 00 sessionSec: " + sessionMSec);
    		int lminutes = 0;
    		
    		/** // requires API level >= 9
//    		lStr = String.format("%d min, %d sec", 
//    			    TimeUnit.MILLISECONDS.toMinutes(sessionMSec),
//    			    TimeUnit.MILLISECONDS.toSeconds(sessionMSec) - 
//    			    TimeUnit.MINUTES.toSeconds(TimeUnit.MILLISECONDS.toMinutes(sessionMSec))
//    			); 
    		 */	
    		lminutes = (int) (sessionMSec / (1000*60));
    		//Log.d(LOGTAG, "longSessionMinInt() 99    minutes: " + lminutes);
    		return lminutes;
    	} // END longSessionMinInt()       
   
    	
    	/**
    	 * Convert date in string ISO 8601 format back to long int milliseconds.
    	 */
    	public static long strDateLong(String dateStr) {
    		//Log.d(LOGTAG, "strDateLong() 00 dateStr: " + dateStr);
    		long ldateLong = 0;
    		Date ldate;
    		SimpleDateFormat iso8601Format = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss"); 		// FIXME: this method will get called frequently -- Formater maybe should be static, or at the object level
    		try {
    			ldate =  iso8601Format.parse(dateStr);
    			ldateLong = ldate.getTime();
    		} catch (java.text.ParseException e) {
    			Log.e(LOGTAG, "strDateLong - Parsing ISO8601 datetime failed", e);
    		}
    		//Log.d(LOGTAG, "strDateLong() 99 dateLong: " + ldateLong);
    		return ldateLong;
    	} // END strDateLong()
    	
    	/**
    	 * Convert session in HH:mm:ss format into milliseconds. 
    	 */
    	public static long strSessionMSecLong(String sessionStr) {
    		//Log.w(LOGTAG, "strSessionMSecLong() 00 sessionStr: " + sessionStr);
    		long ldateLong = 0;
    		Date ldate;
    		SimpleDateFormat sessionFormat = new SimpleDateFormat("HH:mm:ss");

    		try {
    			sessionFormat.setTimeZone(TimeZone.getTimeZone("UTC"));
    			ldate = sessionFormat.parse(sessionStr);
    			ldateLong = ldate.getTime();
    		} catch (java.text.ParseException e) {
    			Log.e(LOGTAG, "strDateLong - Parsing time failed", e);
    		}
    		//Log.w(LOGTAG, "strSessionMSecLong() 99 dateLong: " + ldateLong);
    		return ldateLong;
    	} // END strSessionMSecLong()       
    	
    	
    } // END SessionDB Contract Class
    
    
    private DBHelper mDBHelper;
    //////////////////////////////////////////////
    // DBHelper code creating and managing the database.
    /**
     * This class helps open, create, and upgrade the database file.
     */
    private static class DBHelper extends SQLiteOpenHelper {
    	private Context mContext;

    	/**
    	 * DBHelper constructor.
    	 */
        DBHelper(Context context) {
            super(context, DB_NAME, null, DB_VERSION);
            mContext =  context;
    		Log.d(LOGTAG, "DBHelper constructor");
        }

        /**
         * Create the SQLite database (and insert sample data).
         * @param db SQLiteDatabase handle
         */
        @Override
        public void onCreate(SQLiteDatabase db) {
    		Log.d(LOGTAG, "DBHelper onCreate() 00");
        	// Create tables for practices, goals, instruments, sync, and sync_crud.
    		// Practices table.
        	final String SQL_CREATE_PRACTICES = "CREATE TABLE " + SessionsDB.PRACTICES_TABLE +
        			"( " + 
        			SessionsDB._ID + " INTEGER PRIMARY KEY, " +
        			SessionsDB.PRACTICE_DATE_STAMP + " INTEGER, " + 
        			SessionsDB.PRACTICE_LENGTH_TOTAL + " INTEGER, " +
        			SessionsDB.PRACTICE_LENGTH_NOISE + " INTEGER, " +
        			SessionsDB.PRACTICE_AUDIO_FILE + " TEXT," +
        			SessionsDB.PRACTICE_AUDIO_FILE_KB + " INTEGER," +
        			SessionsDB.PRACTICE_GOAL_ID + " INTEGER, " + 
        			SessionsDB.PRACTICE_NOTES + " TEXT, " +
        			SessionsDB.PRACTICE_INSTRUMENT_ID + " INTEGER, " +
        			SessionsDB.PRACTICE_DATA_VER + " INTEGER, " + 
        			"CONSTRAINT fkgid FOREIGN KEY ( " + SessionsDB.PRACTICE_GOAL_ID + " ) REFERENCES " + SessionsDB.GOALS_TABLE + " ( _id ) ON DELETE SET NULL, " + 
        			"CONSTRAINT fkiid FOREIGN KEY ( " + SessionsDB.PRACTICE_INSTRUMENT_ID + " ) REFERENCES " + SessionsDB.INSTRUMENTS_TABLE + " ( _id ) ON DELETE SET NULL, " +
        			"CONSTRAINT uaf UNIQUE ( " + SessionsDB.PRACTICE_AUDIO_FILE + " ) " +
        			" ) " ;
        	Log.d(LOGTAG, "DBHelper.onCreate() 10 SQL pract: " + SQL_CREATE_PRACTICES);
        	db.execSQL(SQL_CREATE_PRACTICES);
            
        	// Goals table.
        	final String SQL_CREATE_GOALS = "CREATE TABLE " + SessionsDB.GOALS_TABLE +
        			"( " + SessionsDB._ID + " INTEGER PRIMARY KEY, " +
        			SessionsDB.GOAL_NAME + " TEXT, " +
        			SessionsDB.GOAL_DESC + " TEXT, " + 
        			SessionsDB.GOAL_DATA_VER + " INTEGER ) " ;
            Log.d(LOGTAG, "DBHelper.onCreate() 20 SQL goal: " + SQL_CREATE_GOALS);
        	db.execSQL(SQL_CREATE_GOALS);
        	
        	// Instruments table.
        	final String SQL_CREATE_INSTRUMENTS = "CREATE TABLE " + SessionsDB.INSTRUMENTS_TABLE +
        			"( " + SessionsDB._ID + " INTEGER PRIMARY KEY, " +
        			SessionsDB.INSTRUMENT_NAME + " TEXT, " +
        			SessionsDB.INSTRUMENT_DESC + " TEXT, " +
        			SessionsDB.INSTRUMENT_DATA_VER + " INTEGER ) " ;
            Log.d(LOGTAG, "DBHelper.onCreate() 20 SQL instruments: " + SQL_CREATE_INSTRUMENTS);
        	db.execSQL(SQL_CREATE_INSTRUMENTS);

        	// Sync table, with meta data on the device's sync status.
        	final String SQL_CREATE_SYNC = "CREATE TABLE " + SessionsDB.SYNC_TABLE +
        			"( " + SessionsDB._ID + " INTEGER PRIMARY KEY, " +
        			SessionsDB.SYNC_USERID + " TEXT, " + 
        			SessionsDB.SYNC_USERPW + " TEXT, " + 
        			SessionsDB.SYNC_DEVICE + " TEXT, " +
        			SessionsDB.SYNC_DBVER + " INTEGER, " +
        			SessionsDB.SYNC_MPQVER + " TEXT, " +
        			SessionsDB.SYNC_MPQVER_INT + " INTEGER, " +
        			SessionsDB.SYNC_DATE_STAMP + " INTEGER " +
        			" ) " ;
            Log.d(LOGTAG, "DBHelper.onCreate() 20 SQL sync: " + SQL_CREATE_SYNC);
        	db.execSQL(SQL_CREATE_SYNC);

        	// Sync_crud table, with CRUD operations to process next sync.
        	final String SQL_CREATE_SYNC_CRUD = "CREATE TABLE " + SessionsDB.SYNC_CRUD_TABLE +
        			"( " + SessionsDB._ID + " INTEGER PRIMARY KEY, " +
        			SessionsDB.SYNC_CRUD_OP + " TEXT, " +
        			SessionsDB.SYNC_CRUD_OBJECT + " TEXT, " +
        			SessionsDB.SYNC_CRUD_ID + " INTEGER " +
        			" ) " ;
            Log.d(LOGTAG, "DBHelper.onCreate() 20 SQL sync crud: " + SQL_CREATE_SYNC_CRUD);
        	db.execSQL(SQL_CREATE_SYNC_CRUD);
        	

            // Insert sync and sample data.
            ContentValues val = new ContentValues();
        	// Calculate 'now' for sync data_ver columns.
    		long lNowSec = java.lang.System.currentTimeMillis();

            // sync data
            val.put(SessionsDB.SYNC_DBVER, DB_VERSION);
            val.put(SessionsDB.SYNC_MPQVER, mContext.getString(R.string.app_version));
            val.put(SessionsDB.SYNC_MPQVER_INT, Integer.valueOf( mContext.getString(R.string.app_version_int)) );
			SharedPreferences lsharedPref = PreferenceManager.getDefaultSharedPreferences(mContext);
			String luseridPref = lsharedPref.getString(mContext.getString(R.string.pref_userid_key), mContext.getString(R.string.pref_userid_default));
            val.put(SessionsDB.SYNC_USERID, luseridPref);
            Log.d(LOGTAG, "DBHelper.onCreate() 30 insert sample data val: " + val.toString());
            db.insert(SessionsDB.SYNC_TABLE, null, val);            			            
            // goals sample data
            val.clear();
            val.put(SessionsDB.GOAL_NAME, "goal1 - blues");
            val.put(SessionsDB.GOAL_DESC, "Twelve bar blues");
            val.put(SessionsDB.GOAL_DATA_VER, lNowSec);
            db.insert(SessionsDB.GOALS_TABLE, null, val);            			
            // instruments sample data
            val.clear();
            val.put(SessionsDB.INSTRUMENT_NAME, "piano");
            val.put(SessionsDB.INSTRUMENT_DESC, "keyboard");
            val.put(SessionsDB.INSTRUMENT_DATA_VER, lNowSec);
            db.insert(SessionsDB.INSTRUMENTS_TABLE, null, val);            			
            val.clear();
            val.put(SessionsDB.INSTRUMENT_NAME, "harmonica");
            val.put(SessionsDB.INSTRUMENT_DESC, "portable");
            val.put(SessionsDB.INSTRUMENT_DATA_VER, lNowSec);
            db.insert(SessionsDB.INSTRUMENTS_TABLE, null, val);   
            
    		Log.d(LOGTAG, "DBHelper onCreate() 99");
        } // END onCreate()

        /**
         * Upgrade SQLite database.
         * FIXME: currently (during early development ~2012) this simply deletes old database data.  Later, this should alter tables, maintaining data.
         * @param db SQLiteDatabase handle
         * @param oldVersion int with previous version of database found
         * @param newVersion int with the new version of the database to upgrade to
         */
        @Override
        public void onUpgrade(SQLiteDatabase db, int oldVersion, int newVersion) {
            Log.w(LOGTAG, "DBHelper.onUpgrade() 00 Upgrading database from version " + oldVersion + " to " + newVersion);
            if (oldVersion < 17) { // Abusively delete and recreate tables and data during early development.
                Log.w(LOGTAG, "DBHelper.onUpgrade() 01 <17 ");
            	db.execSQL("DROP TABLE IF EXISTS " + SessionsDB.PRACTICES_TABLE);
            	db.execSQL("DROP TABLE IF EXISTS " + SessionsDB.GOALS_TABLE);
            	db.execSQL("DROP TABLE IF EXISTS " + SessionsDB.INSTRUMENTS_TABLE);
            	onCreate(db);
            }
            if (oldVersion < 18) {
                Log.w(LOGTAG, "DBHelper.onUpgrade() 02 <18 ");
            	/**
            	 * MPQ v0.5.6+ 
            	 * PRACTICES updates: (rename PRACTICES, re-create, copy back data)
            	 *   Add FKs to GID and IID
            	 *   Add col PRACTICES.AUDIO_FILE_KB
            	 */  
            	// Rename PRACTICES to a temporary table used during the upgrade.
            	db.execSQL("DROP TABLE IF EXISTS utemp");
            	
            	db.execSQL("ALTER TABLE " + SessionsDB.PRACTICES_TABLE + " RENAME TO utemp");
            	// Create new PRACTICES table w/ FK and new col.
            	String lsql = "CREATE TABLE " + SessionsDB.PRACTICES_TABLE +
            			"( " + 
            			SessionsDB._ID + " INTEGER PRIMARY KEY, " +
            			SessionsDB.PRACTICE_DATE_STAMP + " INTEGER, " + 
            			SessionsDB.PRACTICE_LENGTH_TOTAL + " INTEGER, " +
            			SessionsDB.PRACTICE_LENGTH_NOISE + " INTEGER, " +
            			SessionsDB.PRACTICE_AUDIO_FILE + " TEXT," +
            			SessionsDB.PRACTICE_AUDIO_FILE_KB + " INTEGER," +
            			SessionsDB.PRACTICE_GOAL_ID + " INTEGER, " + 
            			SessionsDB.PRACTICE_NOTES + " TEXT, " +
            			SessionsDB.PRACTICE_INSTRUMENT_ID + " INTEGER, " +
	        			"CONSTRAINT fkgid FOREIGN KEY ( " + SessionsDB.PRACTICE_GOAL_ID + " ) REFERENCES " + SessionsDB.GOALS_TABLE + " ( _id ) ON DELETE SET NULL, " + 
	        			"CONSTRAINT fkiid FOREIGN KEY ( " + SessionsDB.PRACTICE_INSTRUMENT_ID + " ) REFERENCES " + SessionsDB.INSTRUMENTS_TABLE + " ( _id ) ON DELETE SET NULL, " +
	        			"CONSTRAINT uaf UNIQUE ( " + SessionsDB.PRACTICE_AUDIO_FILE + " ) " +
            			" ) " ;
            	Log.w(LOGTAG, "DBHelper.onUpgrade() 10 SQL pract: " + lsql);
            	db.execSQL(lsql);
            	// Copy back data from temporary table.
            	lsql = new String();
            	lsql = lsql + " INSERT INTO " + SessionsDB.PRACTICES_TABLE + " ";
            	lsql = lsql + "  ( " + SessionsDB._ID + " , " +
            			SessionsDB.PRACTICE_DATE_STAMP + " , " + 
            			SessionsDB.PRACTICE_LENGTH_TOTAL + " , " +
            			SessionsDB.PRACTICE_LENGTH_NOISE + " , " +
            			SessionsDB.PRACTICE_AUDIO_FILE + " ," +
            			SessionsDB.PRACTICE_AUDIO_FILE_KB + " ," +
            			SessionsDB.PRACTICE_GOAL_ID + " , " + 
            			SessionsDB.PRACTICE_NOTES + " , " +
            			SessionsDB.PRACTICE_INSTRUMENT_ID + " )  " +
            			" SELECT " + SessionsDB._ID + " , " +
                    			SessionsDB.PRACTICE_DATE_STAMP + " , " + 
                    			SessionsDB.PRACTICE_LENGTH_TOTAL + " , " +
                    			SessionsDB.PRACTICE_LENGTH_NOISE + " , " +
                    			SessionsDB.PRACTICE_AUDIO_FILE + " , " +
                    			" NULL , " +
                    			SessionsDB.PRACTICE_GOAL_ID + " , " + 
                    			SessionsDB.PRACTICE_NOTES + " , " +
                    			SessionsDB.PRACTICE_INSTRUMENT_ID + " " +
                    	" FROM utemp "; 
            	Log.w(LOGTAG, "DBHelper.onUpgrade() 20 SQL pract copy: " + lsql);
            	db.execSQL(lsql);
            	// Leave old data around for now, it's small in size.
            	//db.execSQL("DROP TABLE IF EXISTS utemp");
            	
            	// Review existing files, and update DB with their info.
            	
            	// Get list of files from DB 
            	// Loop through, 
            	//   open with File
            	//   Update DB 
            	Cursor lcursor = db.query(SessionsDB.PRACTICES_TABLE, new String[] {SessionsDB._ID, SessionsDB.PRACTICE_AUDIO_FILE}, null, null, null, null, null);
            	if (null != lcursor) {
            		int lrows = lcursor.getCount();
        			lcursor.moveToFirst();
            		Log.w(LOGTAG, "DBHelper.onUpgrade() 30 file update cursor: " + lrows);
            		long lpid;
            		String lpfilename;
            		long lpfileKB;
            		File lpfile;
        			ContentValues lnewValues;
        			int lrowCount;
            		
            		for (int i = 0; i < lrows; i++) { // FIXME: possibly use an iterator
            			lpid = lcursor.getLong(0);
            			lpfilename = lcursor.getString(1);
                		Log.w(LOGTAG, "DBHelper.onUpgrade() 32 session id: " + lpid + " filename: " + lpfilename);

            			lpfile = new File(Environment.getExternalStorageDirectory().getAbsolutePath() + lpfilename);
            			lpfileKB = lpfile.length() / 1024;  // Convert from bytes, to KB.
            			
            			// Update DB with new length value.
            			lnewValues = new ContentValues();			
            			lnewValues.put(SessionsDB.PRACTICE_AUDIO_FILE_KB, lpfileKB);
            			
            			Log.d(LOGTAG, "DBHelper.onUpgrade() 34 id: " + lpid + " newValues:" + lnewValues.toString() );
            			lrowCount = db.update(SessionsDB.PRACTICES_TABLE, lnewValues, " ( _ID = " + lpid + " ) ", null);
            			Log.w(LOGTAG, "DBHelper.onUpgrade() 35 updated rowcount: " + lrowCount );
            			
            			lcursor.moveToNext();
            		} // END for()
            		
            	} else {
            		Log.e(LOGTAG, "DBHelper.onUpgrade() 39 no cursor to udpate files");
            	}
            } // END if (oldVersion < 18)
            if (oldVersion < 19) {
                Log.w(LOGTAG, "DBHelper.onUpgrade() 39zzz <19 ");
            	/**
            	 * MPQ v0.5.7+ 
            	 * SYNC tables and columns added
            	 *   Add and populate columns for data_version to MPQ tables
            	 *   Add tables Sync and Sync_Crud
            	 *   Populate Sync with basic device info
            	 */  
            	String lsql;
            	// Alter practice, goals, instruments adding sync meta data column.
            	try { // FIXME: hack catch of errors; better to create a new table and copy over, or check PRAGMA table_info(practices).
	            	lsql = "ALTER TABLE " + SessionsDB.PRACTICES_TABLE + 
	            			" ADD COLUMN " + SessionsDB.PRACTICE_DATA_VER + " INTEGER ";
	                Log.d(LOGTAG, "DBHelper.onUpgrade() 40aa alter practice : " + lsql);
	            	db.execSQL(lsql);
	            	lsql = "ALTER TABLE " + SessionsDB.GOALS_TABLE + 
	            			" ADD COLUMN " + SessionsDB.GOAL_DATA_VER + " INTEGER ";
	                Log.d(LOGTAG, "DBHelper.onUpgrade() 40bb alter goals : " + lsql);
	            	db.execSQL(lsql);
	            	lsql = "ALTER TABLE " + SessionsDB.INSTRUMENTS_TABLE + 
	            			" ADD COLUMN " + SessionsDB.INSTRUMENT_DATA_VER + " INTEGER ";
	                Log.d(LOGTAG, "DBHelper.onUpgrade() 40cc alter goals : " + lsql);
	            	db.execSQL(lsql);

	            	// Populate now sync meta data columns.
		    		long lNowSec = java.lang.System.currentTimeMillis();
		        	lsql = "UPDATE " + SessionsDB.PRACTICES_TABLE + 
		        			" SET " + SessionsDB.PRACTICE_DATA_VER + " = " + lNowSec;
		            Log.d(LOGTAG, "DBHelper.onUpgrade() 40mm update practice : " + lsql);
		        	db.execSQL(lsql);
		        	lsql = "UPDATE " + SessionsDB.GOALS_TABLE + 
		        			" SET " + SessionsDB.GOAL_DATA_VER + " = " + lNowSec;
		            Log.d(LOGTAG, "DBHelper.onUpgrade() 40nn update goals : " + lsql);
		        	db.execSQL(lsql);
		        	lsql = "UPDATE " + SessionsDB.INSTRUMENTS_TABLE + 
		        			" SET " + SessionsDB.INSTRUMENT_DATA_VER + " = " + lNowSec;
		            Log.d(LOGTAG, "DBHelper.onUpgrade() 40oo alter goals : " + lsql);
		        	db.execSQL(lsql);
            	} catch(Exception e) {
            		Log.e(LOGTAG, "Eerror: DBHelper.onUpgrade() 40xxx alter table exception : " + e.getLocalizedMessage());
            	}
            	
            	// Create new Sync table, with meta data on the device's sync status.
            	db.execSQL("DROP TABLE IF EXISTS " + SessionsDB.SYNC_TABLE);
            	lsql = "CREATE TABLE " + SessionsDB.SYNC_TABLE +
            			"( " + SessionsDB._ID + " INTEGER PRIMARY KEY, " +
            			SessionsDB.SYNC_USERID + " TEXT, " + 
            			SessionsDB.SYNC_USERPW + " TEXT, " + 
            			SessionsDB.SYNC_DEVICE + " TEXT, " +
            			SessionsDB.SYNC_DBVER + " INTEGER, " +
            			SessionsDB.SYNC_MPQVER + " TEXT, " +
            			SessionsDB.SYNC_MPQVER_INT + " INTEGER, " +
            			SessionsDB.SYNC_DATE_STAMP + " INTEGER " +
            			" ) " ;
                Log.d(LOGTAG, "DBHelper.onUpgrade() 40ZZ updgrade to v19 sync SQL: " + lsql);
            	db.execSQL(lsql);
            	// Create new Sync_crud table, with CRUD operations to process next sync.
            	db.execSQL("DROP TABLE IF EXISTS " + SessionsDB.SYNC_CRUD_TABLE);
            	lsql = new String();
            	lsql = lsql + "CREATE TABLE " + SessionsDB.SYNC_CRUD_TABLE +
            			"( " + SessionsDB._ID + " INTEGER PRIMARY KEY, " +
            			SessionsDB.SYNC_CRUD_OP + " TEXT, " +
            			SessionsDB.SYNC_CRUD_OBJECT + " TEXT, " +
            			SessionsDB.SYNC_CRUD_ID + " INTEGER " +
            			" ) " ;
                Log.d(LOGTAG, "DBHelper.onUpgrade() 41 updgrade to v19 sync_crud SQL: " + lsql);
            	db.execSQL(lsql);
            	
                // Populate sync table.
                ContentValues val = new ContentValues();
                val.put(SessionsDB.SYNC_DBVER, DB_VERSION);
                val.put(SessionsDB.SYNC_MPQVER, mContext.getString(R.string.app_version));
                val.put(SessionsDB.SYNC_MPQVER_INT, Integer.valueOf( mContext.getString(R.string.app_version_int)) );
    			SharedPreferences lsharedPref = PreferenceManager.getDefaultSharedPreferences(mContext);
    			String luseridPref = lsharedPref.getString(mContext.getString(R.string.pref_userid_key), mContext.getString(R.string.pref_userid_default));
                val.put(SessionsDB.SYNC_USERID, luseridPref);
                Log.d(LOGTAG, "DBHelper.onUpgrade() 43 updgrade to v19 insert sample data val: " + val.toString());
                long lnewRowID = db.insert(SessionsDB.SYNC_TABLE, null, val);            			            
                Log.w(LOGTAG, "DBHelper.onUpgrade() 49 updgrade to v19 insert rowid: " + lnewRowID);
            } // END if (oldVersion < 19)
            
        } // END onUpgrade()
        
        /**
         * Enable foreign keys when opening a DB connection.
         * credit: SO Phil -- http://stackoverflow.com/questions/2545558/foreign-key-constraints-in-android-using-sqlite-on-delete-cascade
         */
        @Override
        public void onOpen(SQLiteDatabase db) {
            super.onOpen(db);
            if (!db.isReadOnly()) {
                // Enable foreign key constraints
                db.execSQL("PRAGMA foreign_keys=ON;");
            }
        }
        
    } // END DBHelper class
    

    
    ///////////////////
    // ContentProvider classes 
    /**
     * Initialize the database helper.
     */
	@Override
	public boolean onCreate() {
        mDBHelper = new DBHelper(getContext());
        return true;
	}

	/**
	 * General query handler, which dispatches the query as appropriate.
     * @param uri Uri corresponding with the query
     * @param projection String[] array of strings table columns to query 
     * @param selection String with selection filter
	 * @param selectionArgs String[] array with arguments to insert into selection filters
	 * @param sortOrder String with the sorting statement
     * @return Cursor with query results which is set On to monitor for notifications
	 * @throws IllegalArgumentException if invalid uri is received
	 */
	@Override
	public Cursor query(Uri uri, String[] projection, String selection,
			String[] selectionArgs, String sortOrder) {
		Log.d(LOGTAG, "privder query() 00 uri:" + uri);

        // Use the UriMatcher to see what kind of query we have and format the db query accordingly
		int urimId = sUriMatcher.match(uri);
        switch (urimId) {
            case URIM_PRACTICES:
            case URIM_PRACTICES_ID:
        		Log.d(LOGTAG, "privder query() 20 practices");
        		return getGeneralQuery(uri, projection, selection, sortOrder, urimId);
            case URIM_PRACTICES_COUNT:
        		Log.d(LOGTAG, "privder query() 22 practices count");
        		return getMetaQuery(uri, projection, selection, selectionArgs, urimId);
            case URIM_PRACTICES_PREV_ID:
            	Log.d(LOGTAG, "privder query() 23a practices prev");
            	return getMetaQuery(uri, projection, selection, selectionArgs, urimId);
            case URIM_PRACTICES_NEXT_ID:
            	Log.d(LOGTAG, "privder query() 23b practices next");
            	return getMetaQuery(uri, projection, selection, selectionArgs, urimId);
            case URIM_PRACTICES_FILE:
            	Log.d(LOGTAG, "privder query() 23c practices file");
            	return getMetaQuery(uri, projection, selection, selectionArgs, urimId);            	
            case URIM_PRACTICES_TOTALS:
            	Log.d(LOGTAG, "privder query() 23d practices totals");
            	return getMetaQuery(uri, projection, selection, selectionArgs, urimId);
            case URIM_PRACTICES_FILE_SUM_KB:
            	Log.d(LOGTAG, "privder query() 23e practices file sum kb");
            	return getMetaQuery(uri, projection, selection, selectionArgs, urimId);
            case URIM_INSTRUMENTS:
        		Log.d(LOGTAG, "privder query() 24 general");
        		return getGeneralQuery(uri, projection, selection, sortOrder, urimId);
            case URIM_GOALS:
        		Log.d(LOGTAG, "privder query() 30");
        		return getGeneralQuery(uri, projection, selection, sortOrder, urimId);
            case URIM_SYNC:
        		Log.d(LOGTAG, "privder query() 40");
        		return getGeneralQuery(uri, projection, selection, sortOrder, urimId);
            default:
        		Log.d(LOGTAG, "privder query() 80");
                throw new IllegalArgumentException("query() Unknown Uri: " + uri);
        } // END switch(urimId)
	} // END query()

	@Override
	public String getType(Uri uri) {
		// TODO Auto-generated method stub
		// vnd.android.cursor.dir/vnd.com.example.provider.table1 # muti-row return
		// vnd.android.cursor.item/vnd.com.example.provider.table1 # single row return
		// FIXME: later - implement getStreamTypes() for file support (if that's even required -- we may access files directly for now...)
		return null;
	}

	/**
	 * General insert handler, adding complete rows of data to Sessions DB tables.
     * @param uri Uri corresponding with the insert
     * @param values ContentValues bundle with new data to insert 
     * @return Uri to the data inserted
     * @throws IllegalArgumentException if invalid uri is received
     * @throws SQLException if there is an error inserting row
	 */
	@Override
	public Uri insert(Uri uri, ContentValues values) {
		Log.d(LOGTAG, "privder insert() 00 uri: " + uri + " :values: " + values);
		
		SQLiteDatabase db = mDBHelper.getWritableDatabase();
		long rowId = -1;	// set rowId to a value that represents invalid/unknown data.
		Uri newUri;

		// Calculate current timestamp, and append to insert statement; used for sync.
		long lNowSec = java.lang.System.currentTimeMillis();

        // Use the UriMatcher to see what kind of query we have and format the db query accordingly
        switch (sUriMatcher.match(uri)) {
            case URIM_GOALS:
            	// FIXME: validate data (fields, content) before putting it into the DB
                values.put(SessionsDB.GOAL_DATA_VER, lNowSec);
                rowId = db.insert(SessionsDB.GOALS_TABLE, null, values);
                newUri = SessionsDB.CONTENT_GOALS_URI;
            	break;
            case URIM_INSTRUMENTS:
            	// FIXME: validate data (fields, content) before putting it into the DB
                values.put(SessionsDB.INSTRUMENT_DATA_VER, lNowSec);
                rowId = db.insert(SessionsDB.INSTRUMENTS_TABLE, null, values);
                newUri = SessionsDB.CONTENT_INSTRUMENTS_URI;
                break;
            case URIM_PRACTICES:
            	// FIXME: validate data...
                values.put(SessionsDB.PRACTICE_DATA_VER, lNowSec);
            	rowId = db.insert(SessionsDB.PRACTICES_TABLE, null, values);
            	newUri = SessionsDB.CONTENT_PRACTICES_URI;
            	break;
            // No insert for SYNC table on purpose.  It is a meta-table, with only one row of data.
            default:
        		Log.d(LOGTAG, "privder insert() 40");
                throw new IllegalArgumentException("insert() Unknown Uri: " + uri);
        } // END switch(sUriMatcher)
        
        if (rowId > 0) {
        	// notify of the table change, using the generic newUri
        	getContext().getContentResolver().notifyChange(newUri, null);
        	// update the newUri to include the rowId, and return it
        	newUri = ContentUris.withAppendedId(newUri, rowId);
    		Log.d(LOGTAG, "privder insert() 54 newUri:" + newUri);
            return newUri;
        }
        throw new SQLException("Failed to insert row into " + uri);
	} // END insert()

	/**
	 * Method to update the database.
     * @param uri Uri corresponding with the update
     * @param values ContentValues bundle with data to update 
     * @param selection String with selection filter
	 * @param selectionArgs String[] array with arguments to insert into selection filters
     * @return int with the number of rows updated
     * @throws IllegalArgumentException if invalid uri is received
	 */
	@Override
	public int update(Uri uri, ContentValues values, String selection,
			String[] selectionArgs) {		
		Log.d(LOGTAG, "privder update() 00 uri: " + uri + " :values: " + values.toString() + ":selection:" + selection);
		
		SQLiteDatabase db = mDBHelper.getWritableDatabase();
		int rowCount = 0;  // Number of rows updated.
		
		// Calculate current timestamp, and append to insert statement; used for sync.
		long lNowSec = java.lang.System.currentTimeMillis();

        // Use the UriMatcher to see what will be updated and format the DB query accordingly
        switch (sUriMatcher.match(uri)) {
            case URIM_PRACTICES_ID:
            	// Create where clause matching ID; append to selection if present.
            	long lid = ContentUris.parseId(uri);
            	String lsel = "( _ID = " + lid + " )";
            	if ( (null != selection) && (selection.length() > 0) ) {
            		selection = selection + " AND " + lsel;
            	} else {
            		selection = lsel;
            	}
        		Log.d(LOGTAG, "privder update() 30   PRACT ID sel: " + selection);
                values.put(SessionsDB.PRACTICE_DATA_VER, lNowSec);
            	rowCount = rowCount + db.update(SessionsDB.PRACTICES_TABLE, values, selection, selectionArgs);             	
            	break;
            case URIM_PRACTICES:
        		Log.d(LOGTAG, "privder update() 33   PRACT UPDATE sel: " + selection);
                values.put(SessionsDB.PRACTICE_DATA_VER, lNowSec);
            	rowCount = rowCount + db.update(SessionsDB.PRACTICES_TABLE, values, selection, selectionArgs);             	
            	break;            	
            case URIM_SYNC:
        		Log.d(LOGTAG, "privder update() 34   SYNC UPDATE sel: " + selection);
                // No data_ver field in the sync table.
            	rowCount = rowCount + db.update(SessionsDB.SYNC_TABLE, values, selection, selectionArgs);             	
            	break;            	
            default:
        		Log.d(LOGTAG, "privder update() 40");
                throw new IllegalArgumentException("update() Un-Handled Uri: " + uri);
        } // END switch(sUriMatcher)
        
        if (rowCount > 0) {
    		Log.d(LOGTAG, "privder update() 50 rowCount:" + rowCount);
        	// notify of the table change, using the generic newUri
        	getContext().getContentResolver().notifyChange(uri, null); // FIXME: confirm if we need to notify on the URI with _ID, without _ID, or both...
        }		
		return rowCount;
	}
	
	/**
	 * Method to delete data from database.
     * @param uri Uri corresponding with the delete
     * @param selection String with selection filter
	 * @param selectionArgs String[] array with arguments to insert into selection filters
     * @return int with the number of rows deleted
     * @throws IllegalArgumentException if invalid uri is received
	 */
	@Override
	public int delete(Uri uri, String selection, String[] selectionArgs) {
		Log.d(LOGTAG, "privder delete() 00 uri: " + uri + " :selection: " + selection);
						
		SQLiteDatabase db = mDBHelper.getWritableDatabase();
		int rowCount = 0;  // Number of rows deleted.
// FIXME: add to SYNC_CRUD TABLE
        // Use the UriMatcher to see what will be deleted and format the db query accordingly
        switch (sUriMatcher.match(uri)) {
            case URIM_GOALS:
                rowCount = rowCount + db.delete(SessionsDB.GOALS_TABLE, selection, selectionArgs);
            	break;
            case URIM_INSTRUMENTS:
                rowCount = rowCount + db.delete(SessionsDB.INSTRUMENTS_TABLE, selection, selectionArgs);
                break;
            case URIM_PRACTICES_ID:
            	// Create where clause matching ID; append to selection if present.
            	long lid = ContentUris.parseId(uri);
            	String lsel = "( _ID = " + lid + " )";
            	if ( (null != selection) && (selection.length() > 0) ) {
            		selection = selection + " AND " + lsel;
            	} else {
            		selection = lsel;
            	}
        		Log.d(LOGTAG, "privder delete() 21   PRACT ID sel: " + selection);
            	// NOTE: pass-through to next clause, URIM_PRACTICES, to do the actual DB delete.
            case URIM_PRACTICES:
        		Log.d(LOGTAG, "privder delete() 25   PRACT");
        		// Get filename from DB, and delete audio file.
           		String lsql = "select " + SessionsDB.PRACTICE_AUDIO_FILE + " from " + SessionsDB.PRACTICES_TABLE + " where " + selection; 
           		Log.d(LOGTAG, "privder delete() 26   lsql: " + lsql);
        		Cursor cursor= db.rawQuery(lsql, null);
                if (cursor == null) {
            		Log.e(LOGTAG, "privder delete() 27 no cursor to get delete filename");
                } else if (!cursor.moveToFirst()) {
            		Log.e(LOGTAG, "privder delete() 28 unable to get to first row - return null");
                    cursor.close();
                } else {
                	String lFilename = cursor.getString(0);
                	cursor.close();
                	Log.d(LOGTAG, "privder delete() 29 to send audio service intent - delete: " + lFilename);
                    	
	           		// Send delete Intent to the audio service.
	    			Intent intent = new Intent(getContext(), MPQAudioService.class);
	    			intent.setAction(MPQAudioService.ACTION_DEL);
					// Pass filename via "extra-bundle"  
		        	Bundle lextraArgs = new Bundle();
		        	lextraArgs.putString(MPQAudioService.ARG_FILENAME, lFilename);
		        	intent.putExtras(lextraArgs);
	    			getContext().startService(intent);
    			}

            	Log.d(LOGTAG, "privder delete() 30 about to delete from DB");
        		// Delete DB data.
            	rowCount = rowCount + db.delete(SessionsDB.PRACTICES_TABLE, selection, selectionArgs);
            	break;
        	// No delete for Sync table, by design.
            default:
        		Log.d(LOGTAG, "privder delete() 40");
                throw new IllegalArgumentException("delete() Unknown Uri: " + uri);
        } // END switch(sUriMatcher)
        
        if (rowCount > 0) {
    		Log.d(LOGTAG, "privder delete() 50 rowCount:" + rowCount);
        	// notify of the table change, using the generic newUri
        	getContext().getContentResolver().notifyChange(uri, null);
        }		
		return rowCount;
	} // END delete()

	// Helper private methods
    /**
     * A general method to handle select queries, with the goal of code re-use.
     * @param uri Uri corresponding with the query
     * @param projection String[] array of strings table columns to query 
     * @param selection String with selection filter
     * @param sortOrder String with ORDER clause
     * @param urimId integer corresponding with the UriMatch of uri (to save resources)
     * @return Cursor with query results which is set On to monitor for notifications
     * @throws IllegalArgumentException if invalid uri is received
     */
    private Cursor getGeneralQuery(Uri uri, String[] projection, String selection, String sortOrder, int urimId) {
		Log.d(LOGTAG, "privder getGeneralQuery() 00 uri: " + uri);
        SQLiteQueryBuilder builder = new SQLiteQueryBuilder();

        switch (urimId) {
        	case URIM_INSTRUMENTS:
                builder.setTables(SessionsDB.INSTRUMENTS_TABLE);
        		Log.d(LOGTAG, "privder getGeneralQuery() 11");
        		break;
        	case URIM_GOALS:
        		builder.setTables(SessionsDB.GOALS_TABLE);
        		Log.d(LOGTAG, "privder getGeneralQuery() 13");
        		break;
        	case URIM_PRACTICES:
        	case URIM_PRACTICES_ID:
        		builder.setTables(SessionsDB.PRACTICES_TABLE);
        		Log.d(LOGTAG, "privder getGeneralQuery() 15");
        		if (URIM_PRACTICES_ID == urimId) {
        			long lrow = ContentUris.parseId(uri);
        			selection = " _ID = " + lrow;
        		}
        		break;
        	case URIM_SYNC:
        		builder.setTables(SessionsDB.SYNC_TABLE);
        		Log.d(LOGTAG, "privder getGeneralQuery() 16 sync");
        		break;
            default:
        		Log.d(LOGTAG, "privder getGeneralQuery() 19");
                throw new IllegalArgumentException("getGeneralQuery() Unknown Uri: " + uri);
        } // END switch(urimId)

        // Build query into SQL for DEBUG ease.
        //@SuppressWarnings("deprecation")
		//String lsql = builder.buildQuery(projection, selection, null, null, null, sortOrder, null);
		//Log.d(LOGTAG, "privder getGeneralQuery() 41 SQL buildQuery:" + lsql);
        Cursor cursor = builder.query(mDBHelper.getReadableDatabase(),
                projection, selection, null, null, null, sortOrder);
        if (cursor == null) {
    		Log.d(LOGTAG, "privder getGeneralQuery() 51 no cursor - return null");
            return null;
        } else if (!cursor.moveToFirst()) {
    		Log.d(LOGTAG, "privder getGeneralQuery() 54 unable to get to first row - return null");
            cursor.close();
            return null;
        }
		// Set to check for table changes, so cursor is "auto"updated.
		cursor.setNotificationUri(getContext().getContentResolver(), uri);
		Log.d(LOGTAG, "privder getGeneralQuery() 99 uri: " + uri);
		return cursor;
    } // END getGeneralQuery()
    
    
    /**
	 * Meta queries, that return information about the table rather than
	 * content data from the table. Includes table count, next and previous row
	 * _ID, and min and max row _ID, ID corresponding to a filename, 
	 * totals for length and noise, and sum of file KB space used on SD.
     * 
     * @param uri Uri corresponding with the query
     * @param projection String[] array of strings table columns to query 
     * @param selection String with selection filter
	 * @param selectionArgs String[] array with arguments to insert into selection filters
     * @param urimId integer corresponding with the UriMatch of uri (to save resources)
     * @return Cursor with query results which is set On to monitor for notifications 
     * @throws IllegalArgumentException if invalid uri is received
     */
    private Cursor getMetaQuery(Uri uri, String[] projection, 
    					String selection, String[] selectionArgs, int urimId) {
		Log.d(LOGTAG, "privder getMetaQuery() 00 uri: " + uri);
		String lsql;
		long lrow;
		String [] largs = null;
		SQLiteDatabase db = mDBHelper.getWritableDatabase();
		
		switch (urimId) {
       	case URIM_PRACTICES_COUNT:
    		lsql = "select count(*), min(_ID), max(_ID) from " + SessionsDB.PRACTICES_TABLE; 
       		Log.d(LOGTAG, "privder getMetaQuery() 15   lsql: " + lsql);
       		break;
       	case URIM_PRACTICES_PREV_ID:
			lrow = ContentUris.parseId(uri);
			largs = new String [] { String.valueOf(lrow) };
    		lsql = "select max(_ID) from " + SessionsDB.PRACTICES_TABLE + " where ( _ID < ? ) ";
       		Log.d(LOGTAG, "privder getMetaQuery() 16   lsql: " + lsql);
       		break;
       	case URIM_PRACTICES_NEXT_ID:
			lrow = ContentUris.parseId(uri);
			largs = new String [] { String.valueOf(lrow) };
    		lsql = "select min(_ID) from " + SessionsDB.PRACTICES_TABLE + " where ( _ID > ? ) ";
       		Log.d(LOGTAG, "privder getMetaQuery() 17   lsql: " + lsql);
       		break;
       	case URIM_PRACTICES_FILE:
       		largs = selectionArgs;
       		lsql = "select _ID from " + SessionsDB.PRACTICES_TABLE + " where ( " + SessionsDB.PRACTICE_AUDIO_FILE + " like ? ) "; 
       		Log.d(LOGTAG, "privder getMetaQuery() 18A   lsql: " + lsql + " largs: " + largs.toString() + " args[0]: " + largs[0]);
      		break;
       	case URIM_PRACTICES_TOTALS:
       		lsql = "select sum( " + SessionsDB.PRACTICE_LENGTH_TOTAL + " ), sum( " + SessionsDB.PRACTICE_LENGTH_NOISE + " ) from " + SessionsDB.PRACTICES_TABLE + " "; 
       		Log.d(LOGTAG, "privder getMetaQuery() 18B   lsql: " + lsql);
      		break;
       	case URIM_PRACTICES_FILE_SUM_KB:
       		lsql = "select sum( " + SessionsDB.PRACTICE_AUDIO_FILE_KB + " ) from " + SessionsDB.PRACTICES_TABLE + " "; 
       		Log.d(LOGTAG, "privder getMetaQuery() 18C   lsql: " + lsql);
      		break;
       	default:
    	   Log.d(LOGTAG, "privder getMetaQuery() 19 - crashing - unknown uri: " + uri);
           throw new IllegalArgumentException("getCountQuery() Unknown Uri: " + uri);
		} // END switch(urimId)  
		
		Cursor cursor= db.rawQuery(lsql, largs);
        if (cursor == null) {
    		Log.d(LOGTAG, "privder getMetaQuery() 51 no cursor - return null");
            return null;
        } else if (!cursor.moveToFirst()) {
    		Log.d(LOGTAG, "privder getMetaQuery() 54 unable to get to first row - return null");
            cursor.close();
            return null;
        }
	return cursor;
    } // END getMetaQuery()


} // END class SessionsDBProvider
